MODELS -- Start of the MODELS Section

  -- START MODEL PTOC
  -- Model of a time overcurrent protection relay
  MODEL PTOC

    -- Input Interface at MODEL level : MODELS -> MODEL source code
    INPUT API  -- Current Phase IA
    INPUT BPI  -- Current Phase IB
    INPUT CPI  -- Current Phase IC
    INPUT APV  -- Phase-to-Ground Voltage VAG
    INPUT BPV  -- Phase-to-Ground Voltage VBG
    INPUT CPV  -- Phase-to-Ground Voltage VCG

    -- Output Interface at MODEL level: MODEL source code -> MODELS
    OUTPUT TripCommand  -- TRIP-Command to the Circuit-Breaker / Switch

    CONST PowerNetFreq {val: 50}  -- POWER FREQUENCY [Hz]
    CONST PowerNetCycle {val: 0.02}  -- Cycle [s]
    CONST SamplingFreq {val: 1000}  -- Sampling Frequency [Hz]
    CONST SamplingDT {val: 0.001}  -- Sampling Rate [s]
    CONST SamplesPerCycle {val: 20}  -- Samples per Cycle

    DATA Ig1Setting {dflt: 1.0e10}  -- Starting threshold in A
    DATA T1Setting {dflt: 0.0} -- Delay time in s
    DATA CBDelayTime {dflt: 0.1} -- Delay time of the circuit-breaker in s
    DATA MeasLocation {dflt: -1.0} -- ID of the measuring location

    CONST CBTIMEENABLED {val: 4.0}
    CONST TIMEENABLED {val: 2.0}
    CONST DISABLED {val: 0.0}

    -- Declaration of all variables
    VAR

      -- = 0 (Closed) : CB keeps closed
      -- = 1 (Open) : CB will be opened
      TripCommand

      -- Current time delay after starting in s
      T1Delay, CBDelay

      -- Starting signal
      T1Enabled, CBEnabled

      -- Sampled values of the last cycle
      APIBuffer [0..(SamplesPerCycle - 1)]
      BPIBuffer [0..(SamplesPerCycle - 1)]
      CPIBuffer [0..(SamplesPerCycle - 1)]

      -- Filter coefficients (Discrete Fourier Analysis)
      CosKoeff [0..(SamplesPerCycle - 1)]
      SinKoeff [0..(SamplesPerCycle - 1)]

      -- Phasors of the phase currents in A
      -- Re=Real part, Im=Imaginary part
      -- Amount in A, Phase in Degree
      APIRe, APIIm
      BPIRe, BPIIm
      CPIRe, CPIIm
      APIAmount, APIPhase
      BPIAmount, BPIPhase
      CPIAmount, CPIPhase
      Omega -- Angular Frequency in Hz

    -- One-time initialization of variables, etc. at simulation start
    INIT

      -- CB closed
      TripCommand := Closed

      for i := 0 to (SamplesPerCycle - 1) do
        APIBuffer [i] := 0.0
        BPIBuffer [i] := 0.0
        CPIBuffer [i] := 0.0
      endfor

      APIRe := 0.0
      APIIm := 0.0
      BPIRe := 0.0
      BPIIm := 0.0
      CPIRe := 0.0
      CPIIm := 0.0
      APIAmount := 0.0
      APIPhase := 0.0
      BPIAmount := 0.0
      BPIPhase := 0.0
      CPIAmount := 0.0
      CPIPhase := 0.0
      T1Delay := 0.0
      T1Enabled := DISABLED
      CBDelay := 0.0
      CBEnabled := DISABLED
      Omega := 2.0 * pi * PowerNetFreq

      -- Calcualtion of DFT Filter Coefficients
      for i := 0 to (SamplesPerCycle - 1) do
        CosKoeff[i] := cos(Omega * SamplingDT * i)
        SinKoeff[i] := sin(Omega * SamplingDT * i)
      endfor

    ENDINIT

    -- Execution at each simulation time step
    EXEC

      -- Shifting the sampled values one time step to the past
      for i := 0 to (SamplesPerCycle - 2) do
        APIBuffer[i] := APIBuffer[i+1]
        BPIBuffer[i] := BPIBuffer[i+1]
        CPIBuffer[i] := CPIBuffer[i+1]
      endfor

      -- Insert the new sampled value
      APIBuffer[SamplesPerCycle - 1] := API
      BPIBuffer[SamplesPerCycle - 1] := BPI
      CPIBuffer[SamplesPerCycle - 1] := CPI

      -- Reset variables
      APIRe := 0.0
      APIIm := 0.0
      BPIRe := 0.0
      BPIIm := 0.0
      CPIRe := 0.0
      CPIIm := 0.0

      -- Processing the DFT for the last cycle
      for i := 0 to (SamplesPerCycle - 1) do
        APIRe := APIRe + APIBuffer[i] * CosKoeff[i]
        APIIm := APIIm + APIBuffer[i] * SinKoeff[i]
        BPIRe := BPIRe + BPIBuffer[i] * CosKoeff[i]
        BPIIm := BPIIm + BPIBuffer[i] * SinKoeff[i]
        CPIRe := CPIRe + CPIBuffer[i] * CosKoeff[i]
        CPIIm := CPIIm + CPIBuffer[i] * SinKoeff[i]
      endfor

      APIRe := APIRe * 2.0 / SamplesPerCycle
      APIIm := APIIm * 2.0 / SamplesPerCycle
      BPIRe := BPIRe * 2.0 / SamplesPerCycle
      BPIIm := BPIIm * 2.0 / SamplesPerCycle
      CPIRe := CPIRe * 2.0 / SamplesPerCycle
      CPIIm := CPIIm * 2.0 / SamplesPerCycle

      -- Peak value of the fundamental frequency
      APIAmount := sqrt((APIRe * APIRe) + (APIIm * APIIm))
      BPIAmount := sqrt((BPIRe * BPIRe) + (BPIIm * BPIIm))
      CPIAmount := sqrt((CPIRe * CPIRe) + (CPIIm * CPIIm))

      -- Starting signal: if YES, start of the time delay
      if ((APIAmount > Ig1Setting) or
          (BPIAmount > Ig1Setting) or
          (CPIAmount > Ig1Setting))    and
         (T1Enabled = DISABLED)
      then
        -- Starting time of the time delay in s
        T1Delay := t
        T1Enabled := TIMEENABLED

        -- Reset time delay of circuit-breaker
        CBDelay := 0.0
        CBEnabled := DISABLED
      endif

      -- Reset starting signal
      if ((APIAmount < Ig1Setting)  and
          (BPIAmount < Ig1Setting)  and
          (CPIAmount < Ig1Setting)) and
         (T1Enabled = TIMEENABLED)  then
        -- Reset time delay
        T1Delay := 0.0
        T1Enabled := DISABLED

        -- Reset also time delay of circuit-breaker
        CBDelay := 0.0
        CBEnabled := DISABLED
      endif

      -- Starting time delay of circuit-breaker
      if ((t - T1Delay) >= T1Setting) and
          (T1Enabled = TIMEENABLED)   and
          (CBEnabled = DISABLED)
      then
        -- Starting time delay of the circuit breaker in s
        CBDelay := t
        CBEnabled := CBTIMEENABLED
      endif

      -- Generate TRIP command
      if (CBEnabled = CBTIMEENABLED) and
         (T1Enabled = TIMEENABLED)   and
         ((t - CBDelay) >= CBDelayTime)
      then
        -- TRIP-Kommando
        TripCommand := Open

        -- Reset also time delay of circuit-breaker
        CBDelay := 0.0
        CBEnabled := DISABLED
		
		write ("ATPDES: TRIP CB", MeasLocation, " at ", t, "ms")
      endif

    ENDEXEC

  ENDMODEL
  -- End of the MODEL implementation
  -- ENDE MODEL PTOC


  -- START USE
  -- START PTOC USE PTOC2
  -- Interface from Measuring Probe Prb 2
  INPUT APV002 {TACS(APV002)}
  INPUT BPV002 {TACS(BPV002)}
  INPUT CPV002 {TACS(CPV002)}
  INPUT API002 {TACS(API002)}
  INPUT BPI002 {TACS(BPI002)}
  INPUT CPI002 {TACS(CPI002)}

  -- PTOC2 TRIP Command
  -- Output Interface at MODELS level: MODELS -> Power Network
  -- TRIP-Command to the Circuit-Breaker / Switch
  OUTPUT TRP2

  -- Start of the MODEL instance (task)
  USE PTOC AS PTOC2

    -- Phase Current A, B, C
    INPUT API := API002
    INPUT BPI := BPI002
    INPUT CPI := CPI002

    INPUT APV := APV002
    INPUT BPV := BPV002
    INPUT CPV := CPV002

    -- TRIP-Command
    OUTPUT TRP2 := TripCommand

    -- Settings
    DATA Ig1Setting := 1000.0
    DATA T1Setting := 0.0
    DATA CBDelayTime := 0.05
    DATA MeasLocation := 2

  ENDUSE
  -- End of the MODEL instance (task)

  -- Recording Signals to .PL4-File
  RECORD
    PTOC2.TripCommand AS TRP2
    PTOC2.T1Enabled AS T1_2
  -- END PTOC USE PTOC2
  -- ENDE USE


  -- START USE
  -- START PTOC USE PTOC3
  -- Interface from Measuring Probe Prb 3
  INPUT API003 {TACS(API003)}
  INPUT BPI003 {TACS(BPI003)}
  INPUT CPI003 {TACS(CPI003)}

  -- PTOC3 TRIP Command
  -- Output Interface at MODELS level: MODELS -> Power Network
  -- TRIP-Command to the Circuit-Breaker / Switch
  OUTPUT TRP3

  -- Start of the MODEL instance (task)
  USE PTOC AS PTOC3

    -- Phase Current A, B, C
    INPUT API := API003
    INPUT BPI := BPI003
    INPUT CPI := CPI003

    INPUT APV := APV002
    INPUT BPV := BPV002
    INPUT CPV := CPV002

    -- TRIP-Command
    OUTPUT TRP3 := TripCommand

    -- Settings
    DATA Ig1Setting := 1000.0
    DATA T1Setting := 0.0
    DATA CBDelayTime := 0.05
    DATA MeasLocation := 3

  ENDUSE
  -- End of the MODEL instance (task)

  -- Recording Signals to .PL4-File
  RECORD
    PTOC3.TripCommand AS TRP3
    PTOC3.T1Enabled AS T1_3
  -- END PTOC USE PTOC3
  -- ENDE USE

ENDMODELS -- End of the MODELS Section
